package com.simualteSmack;

import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadFactory;

import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;
import org.xmlpull.v1.XmlPullParserFactory;

import com.simualteSmack.Connection.ListenerWrapper;
import com.simualteSmack.model.Packet;

/**
 * @date 2011-7-31
 * @author zhanghaitao
 * @version 1.0
 * 
 */
public class PacketReader {
	private ExecutorService listenerExecutor;
	private boolean done;
	private XMPPConnection connection;
	private XmlPullParser parser;
	private Thread readerThread;

	protected PacketReader(final XMPPConnection connection) {
		this.connection = connection;
		this.init();
	}

	/**
	 * Initializes the reader in order to be used. The reader is initialized
	 * during the first connection and when reconnecting due to an abruptly
	 * disconnection.
	 */
	protected void init() {
		done = false;

		readerThread = new Thread() {
			public void run() {
				parsePackets(this);
			}
		};

		readerThread.setName("Smack Packet Reader ");
		readerThread.setDaemon(true);

		// create an executor to deliver incoming packets to listeners.
		// we will use a single thread with an unbounded queue.
		listenerExecutor = Executors
				.newSingleThreadExecutor(new ThreadFactory() {

					@Override
					public Thread newThread(Runnable r) {
						Thread thread = new Thread(r,
								"smack listener processor");
						thread.setDaemon(true);
						return thread;
					}
				});
		resetParser();
	}

	/**
	 * Starts the packet reader thread and returns once a connection to the
	 * server has been established. A connection will be attempted for a maximum
	 * of five seconds. An XMPPException will be thrown if the connection fails.
	 * 
	 */
	public void startup() {
		readerThread.start();
	}

	/**
	 * Shuts the packet reader down.
	 */
	public void shutdown() {
		done = true;
		// Shut down the listener executor.
		listenerExecutor.shutdown();
	}

	private void resetParser() {
		try {
			parser = XmlPullParserFactory.newInstance().newPullParser();
			parser.setFeature(XmlPullParser.FEATURE_PROCESS_NAMESPACES, true);
			parser.setInput(connection.reader);
		} catch (XmlPullParserException xppe) {
			xppe.printStackTrace();
		}
	}

	/**
	 * Parse top-level packets in order to process them further.
	 * 
	 * @param thread
	 *            the thread that is being used by the reader to parse incoming
	 *            packets.
	 */
	private void parsePackets(Thread thread) {
		try {
			int eventType = parser.getEventType();
			do {
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("message")) {
						processPacket(PacketParserUtils.parseMessage(parser));
					}
					System.out.println("START_TAG");
				} else if (eventType == XmlPullParser.END_TAG) {
					System.out.println("END_TAG");
				}
				eventType = parser.next();
			} while (!done && eventType != XmlPullParser.END_DOCUMENT
					&& thread == readerThread);
		} catch (Exception e) {
			e.printStackTrace();
			if (!done) {
			}
		}
	}

	private void processPacket(Packet packet) {
		if (packet == null) {
			return;
		}

		// Loop through all collectors and notify the appropriate ones.
		for (PacketCollector collector : connection.getPacketCollectors()) {
			collector.processPacket(packet);
		}

		// Deliver the incoming packet to listeners.
		listenerExecutor.submit(new ListenerNotification(packet));
	}

	/**
	 * A runnable to notify all listeners of a packet.
	 */
	private class ListenerNotification implements Runnable {

		private Packet packet;

		public ListenerNotification(Packet packet) {
			this.packet = packet;
		}

		public void run() {
			for (ListenerWrapper listenerWrapper : connection.recvListeners
					.values()) {
				listenerWrapper.notifyListener(packet);
			}
		}
	}

}
